#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <string.h>
#include "err.h"
#include "db.h"

#define MAX_SIZE 	10240

db_t *db_init (char *name, char *user, char *pass) {
	int i;
	db_t *db = (db_t *) malloc (sizeof (db_t));

#if MYSQL_VERSION_ID >= 32000
	if (!mysql_init (&(db->conn))) {
		err_report (E_DB, "db_init: could not initialize connection");
		free (db);
		return NULL;
	}
#endif
		
	db->connected = 0;
	db->name = name ? strdup (name) : NULL;
	db->user = user ? strdup (user) : NULL;
	db->pass = pass ? strdup (pass) : NULL;

	for (i = 0; i < MAX_RES; i++)
		db->has_result[i] = 0;

	return db;
}

int db_connect (db_t *db) {
	int i = 0;

	if (!db) {
		err_report (E_PARAM, "db_connect: invalid db connection");
		return 0;
	}

	if (db->connected)
		return 1;

	while (i < MAX_RETRY) {
#if MYSQL_VERSION_ID >= 32200
	  if (mysql_real_connect (&(db->conn), NULL, db->user, db->pass, db->name, 0, NULL, 0)) {
#else
 	 if (mysql_connect (&(db->conn), NULL, db->user, db->pass) && !mysql_select_db (&(db->conn), db->name)) {
#endif
			db->connected = 1;
			return 1;
		}

		sleep (1);
		i++;
	}

	err_report (E_DB, "db_connect: %s\n", mysql_error (&(db->conn)));
	return 0;
}

int db_query (db_t *db, int res_idx, char *sql, ...) {
	char buf[MAX_SIZE];
	va_list ap;

	if (!db || !sql) {
		err_report (E_PARAM, "db_query: database or query not given");
		return 0;
	}

	if ((res_idx < 0) || (res_idx >= MAX_RES)) {
		err_report (E_PARAM, "db_query: result index out of bounds");
		return 0;
	}

	if (!db->connected) 
		db_connect (db);

	if (!db->connected) {
		err_report (E_DB, "db_query: database is not connected");
		return 0;
	}
		
	db_release (db, res_idx);

	va_start (ap, sql);
	vsprintf (buf, sql, ap);
	va_end (ap);

	if (mysql_query (&(db->conn), buf)) {
		if (strstr (mysql_error (&(db->conn)), "away") ||
			  strstr (mysql_error (&(db->conn)), "connect")) {
			mysql_close (&(db->conn));
			db->connected = 0;
			if (!db_connect (db) || (mysql_query (&(db->conn), buf))) {
				err_report (E_DB, "db_query: %s", mysql_error (&(db->conn)));
				return 0;
			}
		} else {
			err_report (E_DB, "db_query: %s", mysql_error (&(db->conn)));
			return 0;
		}
	}
	
	if (strncmp (buf, "select", 6))
		return 1;

	if (!(db->res[res_idx] = mysql_store_result (&(db->conn)))) {
#if MYSQL_VERSION_ID >= 32224
		if (mysql_field_count (&(db->conn))) {
#else
		if (mysql_num_fields (&(db->conn))) {
#endif
			err_report (E_DB, "%s", mysql_error (&(db->conn)));
			return 0;
		} else
			return 0;  /* however, here it is not an error.. (hm?) */
	}

	db->has_result[res_idx] = 1;
	return 1;
}

int db_next_result (db_t *db, int res_idx) {
	if (!db) {
		err_report (E_PARAM, "db_next_result: invalid db connection");
		return 0;
	}

	if ((res_idx < 0) || (res_idx >= MAX_RES)) {
		err_report (E_PARAM, "db_next_result: result index out of bounds");
		return 0;
	}

	if (!db->has_result[res_idx] || !db->res[res_idx]) 
		return 0;

	if ((db->row[res_idx] = mysql_fetch_row (db->res[res_idx]))) 
		return 1;

	return 0;
}

char *db_value (db_t *db, int res_idx, int column) {
	if (!db) {
		err_report (E_PARAM, "db_value: invalid db connection");
		return NULL;
	}

	if ((res_idx < 0) || (res_idx >= MAX_RES)) {
		err_report (E_PARAM, "db_query: result index out of bounds");
		return 0;
	}

	if (!db->row[res_idx]) 
		db_next_result (db, res_idx);

	if (!db->row) 
		return NULL;

	return (char *)((db->row[res_idx])[column]);
}

int db_release (db_t *db, int res_idx) {
	if (db->has_result[res_idx]) {
		if (db->res[res_idx])
			mysql_free_result (db->res[res_idx]);
	}
	db->has_result[res_idx] = 0;
	return 1;
}

int db_free (db_t *db) {
	int i;

	if (!db) 
		return 1;

	for (i = 0; i < MAX_RES; i++) 
		if (db->has_result[i] && db->res[i])
			mysql_free_result (db->res[i]);

	if (db->connected)
		mysql_close (&(db->conn));

	free (db);
	return 1;
}
