#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/time.h>
#include <dirent.h>
#include "cfg.h"
#include "err.h"
#include "db.h"
#include "htdb.h"

db_t *db;

int htdb_init (char *db_name, char *db_user, char *db_pass, char *cfg_name) {
	int cfg_id = 0;

	if (!(db = db_init (db_name, db_user, db_pass)) || !db_connect (db)) 
		return 0;

	if (!db_query (db, 0, "select id from htdig_cfg limit 1", cfg_name)) {
		printf ("(notice) database '%s' has no config table, trying to initialize.. ", db_name);
		fflush (stdout);

		if (!db_query (db, 0, "create table htdig_cfg (id int(11) not null auto_increment, name varchar(64), index_size int(11), index_updated int(11), username varchar(8), passwd varchar(16), cron_str varchar(32), cron_updated int(11), primary key(id))"))
			return 0;

		printf ("ok!\n"); fflush (stdout);
	}

	if (!db_query (db, 0, "select cfg_id from htdig_cfg_data limit 1")) {
		printf ("(notice) database '%s' has no data table, trying to initialize.. ", db_name);
		fflush (stdout);

		if (!db_query (db, 0, "create table htdig_cfg_data (cfg_id int(11), name varchar(64), value varchar(255), updated int(11))")) 
			return 0;

		printf ("ok!\n"); fflush (stdout);
	}

	if (!db_query (db, 0, "select id from htdig_cfg where name = '%s'", cfg_name)) 
		return 0;

	if (!db_next_result (db, 0)) {
		printf ("(notice) config '%s' not found in database, trying to create.. ", cfg_name);
		fflush (stdout);

		if (!db_query (db, 0, "insert into htdig_cfg (id, name) values (0, '%s')", cfg_name)) 
			return 0;

		printf ("ok!\n"); fflush (stdout);
	}
		
	if (!db_query (db, 0, "select id from htdig_cfg where name = '%s'", cfg_name) ||
	    !db_next_result (db, 0))
		return 0;

	if (!(cfg_id = atoi (db_value (db, 0, 0)))) 
		err_report (E_DB, "could not determine a valid id for configuration, try again?");

	return cfg_id;
}

htdb_t htdb_configsync (int htdb_id, time_t last_modified) {
	cfg_t *cfg;
	char *name_esc = (char *) alloca (MAX_SIZE);
	char *value_esc = (char *) alloca (MAX_SIZE);
	htdb_t htdb_stats;

	htdb_stats.file2db_count = 0;
	htdb_stats.db2file_count = 0;
	htdb_stats.index_size = 0;
	htdb_stats.index_updated = 0;
	htdb_stats.cron_res = 0;

	if (!htdb_id) {
		err_report (E_PARAM, "no valid id given");
		return htdb_stats;
	}

	printf ("(notice) configfile was last modified on %s", ctime(&last_modified));
	fflush (stdout);

	for (cfg = cfg_first(); cfg; cfg = cfg->next) {
#if MYSQL_VERSION_ID >= 32232
		mysql_real_escape_string (&(db->conn), name_esc, cfg->name, strlen (cfg->name));
		mysql_real_escape_string (&(db->conn), value_esc, cfg->value, strlen (cfg->value));
#else
		mysql_escape_string (name_esc, cfg->name, strlen (cfg->name));
		mysql_escape_string (value_esc, cfg->value, strlen (cfg->value));
#endif
		if (!db_query (db, 0, "select value, updated from htdig_cfg_data where cfg_id = %i and name = '%s'", htdb_id, name_esc)) {
			printf ("htdb_sync: warning: %s\n", err_get());
			err_discard();
		} else {
			if (db_next_result (db, 0)) {
				if (atoi (db_value (db, 0, 1)) > last_modified) {
					free (cfg->value);
					cfg->value = strdup (db_value (db, 0, 0));
					htdb_stats.db2file_count++;
				} else {
					if (!db_query (db, 0, "update htdig_cfg_data set value = '%s', updated = %i where name = '%s' and cfg_id = %i", value_esc, time(NULL), name_esc, htdb_id)) {	
						printf ("(notice) while updating %s: %s\n", cfg->name, err_get());
						err_discard();
					} else
						htdb_stats.file2db_count++;
				}
			} else {
				if (!db_query (db, 0, "insert into htdig_cfg_data (cfg_id, name, value, updated) values (%i, '%s', '%s', %i)", htdb_id, name_esc, value_esc, time(NULL))) {	
					printf ("(notice) while inserting %s: %s\n", cfg->name, err_get());
					err_discard();
				} else
					htdb_stats.file2db_count++;
			}
		}
	}

	if (!db_query (db, 0, "select name, value from htdig_cfg_data where cfg_id = %i", htdb_id)) 
		return htdb_stats;

	err_discard();
	while (db_next_result (db, 0)) {
		cfg_retr (db_value (db, 0, 0));
		if (err_getno() != E_NONE) {
			cfg_add (db_value (db, 0, 0), db_value (db, 0, 1));
			err_discard();
			htdb_stats.db2file_count++;
		}
	}

	return htdb_stats;
}

htdb_t htdb_indexsync (int htdb_id) {
	char *buf = (char *) alloca (MAX_SIZE);
	char *buf2 = (char *) alloca (MAX_SIZE);
	char *db_dir;
	char *cfg_var;
	DIR *dir;
	struct dirent *dir_entry;
	struct stat dir_entry_info;
	int i = 0, j, k = 0;
	htdb_t htdb_stats;

	htdb_stats.file2db_count = 0;
	htdb_stats.db2file_count = 0;
	htdb_stats.index_size = 0;
	htdb_stats.index_updated = 0;
	htdb_stats.cron_res = 0;

	if (!htdb_id) {
		err_report (E_PARAM, "no valid id given");
		return htdb_stats;
	}

	if (!strlen (db_dir = cfg_retr ("database_dir"))) {
		err_report (E_FILE, "could not determine database_dir from configfile");
		return htdb_stats;
	}

	while (db_dir[i] && db_dir[i + 1] &&
	       (db_dir[i] != '$') && (db_dir[i + 1] != '{')) {
		buf[i] = db_dir[i];
		i++;
	}
	k = i;
		
	if (db_dir[i] == '$') {
		j = i + 2;
		while (db_dir[j] && (db_dir[j] != '}')) {
			buf2[j - i - 2] = db_dir[j];
			j++;
		}
		buf2[j - i - 2] = 0;
		if (!strlen (cfg_var = cfg_retr (buf2))) {
			err_report (E_FILE, "variable %s in 'database_dir' was not defined", buf2);
			return htdb_stats;
		}
		i = j + 1;
		j = 0;
		while (cfg_var[j]) {
			buf[k] = cfg_var[j];
			k++; j++;
		}
	}

	while (db_dir[i]) {
		buf[k] = db_dir[i];
		k++;
		i++;
	}
	buf[k] = 0;

        if (!(dir = opendir (buf))) {
		err_report (E_FILE, "database_dir '%s' cannot be accessed", buf);
                return htdb_stats;
	}

	printf ("(notice) indexes are stored in '%s'\n", buf);
	fflush (stdout);

	rewinddir (dir);
	while ((dir_entry = readdir (dir))) {
		if (buf[strlen (buf) - 1] != '/')
			sprintf (buf2, "%s/%s", buf, dir_entry->d_name);
		else
			sprintf (buf2, "%s%s", buf, dir_entry->d_name);

		if (!stat (buf2, &dir_entry_info) &&
		    (dir_entry_info.st_mode & S_IFREG)) {	
			htdb_stats.index_size += dir_entry_info.st_size;
			if (dir_entry_info.st_mtime > htdb_stats.index_updated)
				htdb_stats.index_updated = dir_entry_info.st_mtime;
		}
	}

	closedir (dir);
			
	htdb_stats.index_size = htdb_stats.index_size / 1024;

	db_query (db, 0, "update htdig_cfg set index_size = %i, index_updated = %i where id = %i", (int)htdb_stats.index_size, (int)htdb_stats.index_updated, htdb_id);

	return htdb_stats;
}

htdb_t htdb_cronsync (int htdb_id, char *cron_user, char *cron_search) {
	FILE *f;
	char *buf = (char *) alloca (MAX_SIZE);
	char *cron_str = NULL;

	htdb_t htdb_stats;

	htdb_stats.file2db_count = 0;
	htdb_stats.db2file_count = 0;
	htdb_stats.index_size = 0;
	htdb_stats.index_updated = 0;
	htdb_stats.cron_res = 0;

	if (!htdb_id) {
		err_report (E_PARAM, "no valid id given");
		return htdb_stats;
	}

	sprintf (buf, "crontab -u %s -l", cron_user);

	if (!(f = popen (buf, "r"))) {
		err_report (E_FILE, "cannot open crontab for reading\n");
		return htdb_stats;
	}

	while (fgets (buf, MAX_SIZE - 1, f)) {
		if (strstr (buf, cron_search))
			cron_str = strdup (buf);
	}

	fclose (f);

	if (!cron_str) {
		err_report (E_FILE, "crontab of %s is not present or has no entry for '%s'", cron_user, cron_search);
		return htdb_stats;
	}

	printf ("(notice) cron entry = %s\n", cron_str);

	db_query (db, 0, "update htdig_cfg set cron_str = '%s', cron_updated = %i where id = %i", cron_str, time(NULL), htdb_id);
	
	htdb_stats.cron_res = 1;

	return htdb_stats;
}
